unit GStreams;
//=== File Prolog ============================================================
// This code was developed by RiverSoftAVG Based
//      on Ray Lischner's Secrets of Delphi 2 TS_BufferStream and
//      TS_TextStream objects
//
//--- Notes ------------------------------------------------------------------
//
//--- Development History  ---------------------------------------------------
//
// 12/2015 T. Grubb
//         - Fixed signature of TGBufferStream.Read method to match changes XE8+
//           did for iOS64 (Integer->LongInt)
//         - Fixed signature of TGBufferStream.Write method to match changes XE8+
//           did for iOS64 (Integer->LongInt)
//         - Fixed signature of TGBufferStream.Seek method to match changes XE8+
//           did for iOS64 (Integer->LongInt)
//         - Fixed signature of TGTextStream.SeekExcludePreamble method to match changes XE8+
//           did for iOS64 (Integer->LongInt)
//         - Fixed signature of TGTextStream.Seek method to match changes XE8+
//           did for iOS64 (Integer->LongInt)
// 04/2014 T. Grubb
//         - Updated code for XE6
//         - Moved vtChar and vtPChar code in TGTextStream.WriteArgs method
//           to not be used by NEXTGEN compiler
// 06/2013 T. Grubb
//         - Removed TGTextStream.Put(AnsiChar) for NEXTGEN compiler
//         - Fixed TGTextStream.WriteArgs method for NEXTGEN compiler
//         - Changed base class of TBaseStreamTokenList from TIList to
//           TInterfaceListPersistent
//         - Modified TBaseStreamToken reference count code for NEXTGEN compiler
// 02/2012 T. Grubb
//         - Added support for FireMonkey by removing Windows/VCL dependencies
// 11/2011 T. Grubb
//         - Fixed Pointer to Integer Casts to use TRSNativeInt
// 12/2008 T. Grubb
//         - Changed behavior of TGTextStream.SetEncoding method to set PreambleStreamed
//           flag to True if new encoding <> nil (Assumption is that if you already
//           know the encoding, you read the preamble or it doesn't exist (e.g., coming
//           from a string control))
//         - Added optional Encoding parameter to TGTextStream constructors
//         - Fixed behavior where TGTextStream detects if it is working with a
//           TStringStream and then uses its encoding
// 11/2008 T. Grubb
//         - Fixed TGTextStream.GetChar method to correctly read 1 character without
//           causing a PutBack overflow exception
//         - Fixed access violation in TGTextStream.GetChar for D2009 by adding
//           check for reading 0 bytes
// 11/2008 T. Grubb
//         - Modified stream classes to work with Unicode and D2009
//         - Modified stream classes to avoid warnings about collisions with
//           new methods added to TObject (D2009/.NET)
//         - Updated TGTextStream.WriteArgs to handle Int64, WideStrings and
//           Unicode strings
//         - Fixed bug in TGBufferStream.FillBuffer which would cause subsequent PutBack
//           call to fail if at end of stream
//         - Added SPreamblePastEof resource string
//         - Changed return type of TGBufferStream.FillBuffer from boolean to
//           integer
//         - Added TGBufferStream.PrepareBuffer method
//         - Added overloaded TGBufferStream.PutBack(Byte) and PutBack(Char/WideChar)
//           methods
//         - Added ability for TGTextStream to read/write non-default encoded
//           files in Delphi 2009 or later
//         - Added protected TGTextStream.PreambleSize method and PreambleStreamed
//           property
//         - Added TGTextStream.Encoding property for Delphi 2009
//         - Changed return type TGTextStream.GetChar method to Char from AnsiChar
//         - Added overloaded TGTextStream.GetChar(Char) method
//         - Added TGTextStream.SeekExcludePreamble method
// 06/2007 T. Grubb
//         - Removed Published keyword from classes to
//           get rid of D2007 warnings
// 12/2005 T. Grubb
//         - Fixed hang bug in TGTextStream.GetChar introduced in Delphi 2006 by
//           initializing result in the GetChar instead of the Read. (Thanks P.G.)
// 08/2005 T. Grubb
//         - Modified properties and added accessor methods of interfaces to
//           use underscore in order to get rid of annoying .NET hint
// 03/2005 T. Grubb
//         - Changed ToString property to AsString and GetToString method to
//           ToString to sync better with .NET (in which all objects have ToString
//           method)
//         - Changed Tag property to be of TRSTag type (Win32=Integer;.NET=Variant)
//         - Changed QueryInterface calls to use Supports function
//         - Introduced TxxxPersistent class to get around a delphi
//           compiler bug with class helpers (QC report #11019)
//         - Added Close method to TGBufferStream (useful for .NET)
// 02/2005 T. Grubb
//         - Changed IUnknown parameters to IInterface
//         - Removed interface reference counting functions for .NET (not needed)
//         - TGSpecialFiler is not available for .NET (because can't crack
//           protected methods of TReader/TWriter)
// 10/2003 T. Grubb
//         - Added TBaseStreamToken class, which implements the IBaseStreamToken
//           interface
//         - Added TBaseStreamTokenList class, which implements the IBaseStreamTokenList
//           interface
// 04/2003 T. Grubb
//         - Added TGSpecialFiler class which allows writing and reading
//           non-homogeneous collections (e.g., collection items can be of
//           different class types
// 05/1997 T. Grubb
//         Initial version
//
//      File Contents:
//           TBaseStreamToken object that represents a token from a stream
//           TGSpecialFiler   class which accepts a writer and uses it to write
//                            heterogeneous collections (collections that contain
//                            items of different item classes.
//           TGBufferStream   Object for Buffering a stream
//           TGTextStream     TGBufferStream descendant for text streams
//
//
//--- Warning ----------------------------------------------------------------
// This software is property of RiverSoftAVG. Unauthorized use or
//      duplication of this software is strictly prohibited. Authorized users
//      are subject to the following restrictions:
//      *   RiverSoftAVG is not responsible for
//          any consequence of the use of this software.
//      *   The origin of this software must not be misrepresented either by
//          explicit claim or by omission.
//      *   Altered versions of this software must be plainly marked as such.
//      * This notice may not be removed or altered.
//
//       1997-2014, Thomas G. Grubb
//
//=== End File Prolog ========================================================

interface

{$INCLUDE RSDefines.inc}

uses
  RSInterfaceCollections,
  SysUtils, Classes, RSConstants, CommonInterfaces;

resourcestring
  SSeekError = 'Error seeking in buffer';
  SInvalidOrigin = 'Invalid Origin Point specified: %s';
  SPutBackOverflow = 'Overflow on push back';
  SSeekPastEof = 'Invalid seek command, EOF reached';   
  SPreamblePastEof = 'Invalid stream encoding determination, EOF reached';
  SInvalidBufferSize = 'Invalid buffer size specified: %s';
  SInvalidBufferState = 'Invalid buffer state';
  SNoReaderDefined = 'No TReader defined for filer';
  SNoWriterDefined = 'No TWriter defined for filer';
  SInvalidBaseTokenHint = 'Invalid Token hint for parsing token: %s';
  SUnbalancedString = 'Invalid string value, string was not terminated';

const
  IndentString: String = '    ';
  WhiteSpaceDelimiters: String = ' '+#13#10+#9;
  PunctuationDelimiters: String = ',./?<>;:"[{]}`~!@#$%^&*()-=+\|''';
  FloatFormatStr = '%.4f';
  SEscapeChar = '\';
  SStringChar = '"';

type
  TBufferState = (bsUnknown, bsRead, bsWrite);
  TBufferType = {$IFDEF CLR} Array of Byte {$ELSE} {$IFDEF UNICODE}PByte{$ELSE}PChar{$ENDIF} {$ENDIF};
  TBufferIndexType = {$IFDEF CLR} Integer {$ELSE} {$IFDEF UNICODE}PByte{$ELSE}PChar{$ENDIF} {$ENDIF};
  TBufferInternalType = {$IFDEF UNICODE}Byte{$ELSE}Char{$ENDIF};

  TGBufferStream = class(TStream)
  { Purpose: To implement a stream which buffers I/O for increased speed.  Based
    on Lischner's Secrets of Delphi 2 TS_BufferStream
    Note: In .NET, you must call Close or Free }
  private
    { Private declarations }
    FStream: TStream;
    FBuffer: TBufferType;
    FBufferPtr: TBufferIndexType;
    FBufferEnd: TBufferIndexType;
    FState: TBufferState;
    FOnFillBuffer: TNotifyEvent;
    FOnFlushBuffer: TNotifyEvent;
    FBufferSize: Cardinal;
    function GetBufferPosition: Integer;
    procedure SetBufferSize(const Value: Cardinal);
  protected
    { Protected declarations }
{$IFDEF CLR}
    procedure SetSize(NewSize: Int64); override;
{$ENDIF}
    function FillBuffer: Integer; dynamic;
    function FlushBuffer: Boolean; dynamic;
    function PrepareBuffer: Boolean; dynamic;
    procedure PutBack(Ch: Byte); overload; virtual;
    procedure PutBack(Ch: Char); overload; virtual;
    property Buffer: TBufferType read FBuffer;
    property BufferPtr: TBufferIndexType read FBufferPtr;
    property BufferEnd: TBufferIndexType read FBufferEnd;
    property BufferPosition: Integer read GetBufferPosition;
    property Stream: TStream read FStream;
  public
    { Public declarations }
    constructor Create(Stream: TStream); virtual;
    destructor Destroy; override;
    procedure Close;
{$IFNDEF CLR}
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
{$ELSE}
    function Read(var Buffer: array of Byte; Offset, Count: Longint): Longint; override;
    function Write(const Buffer: array of Byte; Offset, Count: Longint): Longint; override;
    function Seek(const Offset: Int64; Origin: TSeekOrigin): Int64; override;
{$ENDIF}
    function IsEof: Boolean;
    property OnFillBuffer: TNotifyEvent read FOnFillBuffer write FOnFillBuffer;
    property OnFlushBuffer: TNotifyEvent read FOnFlushBuffer write FOnFlushBuffer;
    property State: TBufferState read FState;
    property BufferSize: Cardinal read FBufferSize write SetBufferSize default 8192;
  end; { TGBufferStream }
  TGBufferStreamClass = class of TGBufferStream;

  TGTextStream = class(TGBufferStream)
  { Purpose: To implement a stream which provides basic text capabilities.  Based
    on Lischner's Secrets of Delphi 2 TS_TextStream.  The OwnStream Property states
    whether we own this stream and should free it on close. }
  private
    { Private declarations }
    FOwnStream: Boolean;
    FPreambleStreamed: Boolean;
    FDelimiters: String;
{$IFDEF DELPHI12PLUS}
    FEncoding: TEncoding;
{$ENDIF}
    procedure SetDelimiters(const Value: String);
  protected
    { Protected declarations }
    procedure FreeStream; virtual;
{$IFDEF DELPHI12PLUS}
    function PrepareBuffer: Boolean; override;
    procedure PutBack(Ch: Char); overload; override;
{$ENDIF}
    function FillBuffer: Integer; override;
    function PreambleSize: Integer; dynamic;
    property OwnStream: Boolean read FOwnStream write FOwnStream;
    property PreambleStreamed: Boolean read FPreambleStreamed write FPreambleStreamed;
  public
    { Public declarations }
    constructor Create(Stream: TStream{$IFDEF DELPHI12PLUS}; Encoding: TEncoding = nil{$ENDIF}); reintroduce; overload; {$IFDEF DELPHI12PLUS}virtual;{$ELSE}override;{$ENDIF}
    constructor Create(const Filename: String; Mode: Word{$IFDEF DELPHI12PLUS}; Encoding: TEncoding = nil{$ENDIF}); reintroduce; overload;
    destructor Destroy; override;
    procedure SkipSpaces;
    function SkipInLineSpaces: Boolean; virtual;
    function GetChar: Char; overload; virtual;
    function GetChar( var AChar: Char ): Boolean; overload; virtual;
    function GetInteger: Integer;
    function GetFloat: Extended;
    function GetToken: String; overload;
    function GetToken(const Delimiters: String; RestoreOld: Boolean): String; overload;
    function GetLine: String; virtual;
{$IFNDEF NEXTGEN}
    function Put(Ch: AnsiChar): TGTextStream; overload;
{$ENDIF !NEXTGEN}
    function Put(WCh: WideChar): TGTextStream; overload;
    function Put(Int: Integer): TGTextStream; overload;
    function Put(Flt: Extended): TGTextStream; overload;
    function Put(const Str: String): TGTextStream; overload;
    function PutLine(const Str: String): TGTextStream; virtual;
    function PutSpace: TGTextStream;
    function PutTab: TGTextStream;
    function PutEOL: TGTextStream; virtual;
{$IFNDEF CLR}
    function Seek(Offset: Longint; Origin: Word): Longint; override;
    function SeekExcludePreamble(Offset: Longint; Origin: Word): Longint;
{$ENDIF}
    procedure WriteArgs(Args: array of const);
    procedure WriteLn(Args: array of const);
    procedure Format(const Fmt: String; Args: array of const);
    procedure FormatLn(const Fmt: String; Args: array of const);
    property Delimiters: String read FDelimiters write SetDelimiters;
{$IFDEF DELPHI12PLUS}
    property Encoding: TEncoding read FEncoding;
{$ENDIF}
  end; { TGTextStream }
  TGTextStreamClass = class of TGTextStream;

  TStreamTokenType = ( sttNone, sttAtom, sttString, sttInteger, sttFloat, sttVariable );
  TStreamTokenTypes = set of TStreamTokenType;
  // TBaseStreamTokenPersistent is a class only introduced to get around a delphi
  // compiler bug with class helpers (QC report #11019)
  TBaseStreamTokenPersistent = class(TPersistent);
  TBaseStreamToken = class(TBaseStreamTokenPersistent, IBaseStreamToken, ICloneable, IInterface)
  { Purpose: To provide a token from a stream that also holds line number }
  private
    { Private declarations }
    FDataType: Integer;
    FLineNumber: Integer;
    FNumber: Double;
    FText: String;
    FTag: TRSTag;
    FWhitespace: String;
{$IFNDEF AUTOREFCOUNT}
    {$IFDEF DELPHI18PLUS}[Volatile] {$ENDIF}FRefCount: Integer;
{$ENDIF}
  protected
    { Protected declarations }
    { IInterface }
    function _AddRef: Integer; stdcall;
    function _Release: Integer; stdcall;
    function QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;
  public
    { Public declarations }
    function Equals( AValue: TPersistent ): Boolean; {$IFDEF TOBJECT_CHANGES}reintroduce;{$ENDIF}overload; virtual;
    { ICloneable }
    procedure Assign( const AValue: IInterface ); reintroduce; overload;
    function Clone: IInterface;
    function Equals( const AValue: IInterface ): Boolean; {$IFDEF TOBJECT_CHANGES}reintroduce;{$ENDIF}overload;
    function GetTag: TRSTag;
    function ICloneable.get_Tag = GetTag;
    procedure SetTag(const Value: TRSTag);
    procedure ICloneable.set_Tag = SetTag;
    { IBaseStreamToken }
{$IFNDEF CLR}
    function IBaseStreamToken.get_Tag = GetTag;
    procedure IBaseStreamToken.set_Tag = SetTag;
{$ENDIF}
    function GetDataType: Integer;
    function IBaseStreamToken.get_DataType = GetDataType;
    function GetLineNumber: Integer;
    function IBaseStreamToken.get_LineNumber = GetLineNumber;
    function GetNumber: Double;
    function IBaseStreamToken.get_Number = GetNumber;
    function GetText: String;
    function IBaseStreamToken.get_Text = GetText;
    function ToString: String; {$IFNDEF TOBJECT_CHANGES}reintroduce; virtual;{$ELSE}override;{$ENDIF}
    function IBaseStreamToken.get_AsString = ToString;
    function GetWhitespace: String;
    function IBaseStreamToken.get_Whitespace = GetWhitespace;
    procedure SetWhitespace(const Value: String);
    procedure IBaseStreamToken.set_Whitespace = SetWhitespace;
    procedure SetDataType(const Value: Integer);
    procedure IBaseStreamToken.set_DataType = SetDataType;
    procedure SetLineNumber(const Value: Integer);
    procedure IBaseStreamToken.set_LineNumber = SetLineNumber;
    procedure SetNumber(const Value: Double);
    procedure IBaseStreamToken.set_Number = SetNumber;
    procedure SetText(const Value: String);
    procedure IBaseStreamToken.set_Text = SetText;

    procedure Assign( Source: TPersistent ); overload; override;
    property AsString: String read ToString;
  published
    { Published declarations }
    property DataType: Integer read FDataType write FDataType;
    property LineNumber: Integer read FLineNumber write FLineNumber;
    property Text: String read FText write FText;
    property Number: Double read FNumber write FNumber;
    property Tag: TRSTag read FTag write FTag;
    property WhiteSpace: String read FWhitespace write FWhitespace;
  end; { TBaseStreamToken }

  TBaseStreamTokenList = class(TIList, IBaseStreamTokenList, ICloneable, IInterface)
  { Purpose: To provide a list of stream tokens }
  private
    { Private declarations }
    FTag: TRSTag;
  protected
    { Protected declarations }
    function GetValueCount: Integer;
  public
    { Public declarations }
    procedure Assign( Source: TPersistent ); reintroduce; overload; virtual;
    procedure Assign( const AValue: IInterface ); reintroduce; overload;
    procedure Append( const List: IBaseStreamTokenList );
    function Clone: IInterface; reintroduce; overload;
    { IBaseStreamTokenList }
    function Get(Index: Integer): IBaseStreamToken; reintroduce; overload;
    function IBaseStreamTokenList.get_Value = Get;
    function IBaseStreamTokenList.get_ValueCount = GetValueCount;
    procedure Put(Index: Integer; const Value: IBaseStreamToken); reintroduce; overload;
    procedure IBaseStreamTokenList.set_Value = Put;
    function Equals( const AValue: IInterface ): Boolean; {$IFDEF TOBJECT_CHANGES}reintroduce;{$ENDIF}overload;
    function Equals( const AList: IBaseStreamTokenList ): Boolean; {$IFDEF TOBJECT_CHANGES}reintroduce;{$ENDIF}overload; virtual;
    function Equals( AList: TObject ): Boolean; {$IFDEF TOBJECT_CHANGES}reintroduce; {$ENDIF}overload; virtual;
    function GetTag: TRSTag;
    function ICloneable.get_Tag = GetTag;
    procedure SetTag( const Value: TRSTag ); virtual;
    procedure ICloneable.set_Tag = SetTag;
{$IFNDEF CLR}
    function IBaseStreamTokenList.get_Tag = GetTag;
    procedure IBaseStreamTokenList.set_Tag = SetTag;
{$ENDIF}
    function ToString: String; {$IFDEF TOBJECT_CHANGES}override{$ELSE}virtual{$ENDIF};
    function IBaseStreamTokenList.get_AsString = ToString;
    function Add(const Item: IBaseStreamToken): Integer; overload;
    function First: IBaseStreamToken;
    function IndexOf(const Item: IBaseStreamToken): Integer;
    procedure Insert(Index: Integer; const Item: IBaseStreamToken);
    function Last: IBaseStreamToken;
    function Remove(const Item: IBaseStreamToken): Integer;
    property ValueCount: Integer read GetValueCount;
    property Value[ Index: Integer ]: IBaseStreamToken read Get write Put; default;
    property AsString: String read ToString;
  published
    { Published declarations }
    property Tag: TRSTag read FTag write SetTag;
  end; { TBaseStreamTokenList }

{$IFNDEF CLR}
  TGSpecialFiler = class
  { Purpose: To define a class which accepts a writer and uses it to write
    heterogeneous collections (collections that contain items of different
    item classes.
    To use:
       For the class that has the heterogeneous collection property.
           - Make the Collection property Stored False
           - Make DefineProperties, LoadXXXProperty and StoreXXXProperty methods
             for the class that owns the collection
    procedure DefineProperties(Filer: TFiler); override;
    begin
         inherited DefineProperties(Filer);
         Filer.DefineProperty( 'HPresets',
                               LoadPresetsProperty, StorePresetsProperty,
                               Presets.Count > 0 );
    end;
    procedure LoadPresetsProperty(Reader: TReader);
    var
       Filer: TGSpecialFiler;
    begin
         Filer := TGSpecialFiler.Create(Reader);
         try
            Filer.ReadCollection(Presets);
         finally
            Filer.Free;
         end;
    end;
    procedure StorePresetsProperty(Writer: TWriter);
    var
       Filer: TGSpecialFiler;
    begin
         Filer := TGSpecialFiler.Create(Writer);
         try
            Filer.WriteCollection(Presets);
         finally
            Filer.Free;
         end;
    end;

    Note: This class can only read and write heterogeneous collections at the
    TOP level.  If a heterogeneous collection contains a heterogeneous collection
    property, you need to defineproperty again
    Note: all collection item classes must be defined with RegisterClass }
  private
    { Private declarations }
    FWriter: TWriter;
    FReader: TReader;
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create; overload; virtual;
    constructor Create(Reader: TReader); overload;
    constructor Create(Writer: TWriter); overload;
    property Reader: TReader read FReader;
    property Writer: TWriter read FWriter;
    procedure ReadCollection( Collection: TCollection );
    procedure WriteCollection( Value: TCollection );
  end; { TGSpecialFiler }
{$ENDIF}

implementation

uses
    {$IFDEF DELPHI6PLUS}Variants,{$ENDIF}
    {$IFNDEF DELPHI16PLUS}Windows,{$ELSE}SyncObjs,{$ENDIF}
    HelperFunctions;

{ TGBufferStream }

constructor TGBufferStream.Create(Stream: TStream);
{ Create and initialize a buffer stream. }
begin
     inherited Create;
     FStream := Stream;
     FBufferSize := 8192;
     { Allocate the buffer. }
{$IFDEF CLR}
     SetLength(FBuffer, FBufferSize);
     FBufferPtr := -1;
     FBufferEnd := FBufferSize;
{$ELSE}
     GetMem(FBuffer, FBufferSize);
     FBufferEnd := Buffer + FBufferSize;
{$ENDIF}
     FState := bsUnknown;
end;

destructor TGBufferStream.Destroy;
{ Destroy the buffer stream, before destroying, write it out if needed. }
begin
     Close;
{$IFNDEF CLR}
     FreeMem(FBuffer, BufferSize);
{$ENDIF}
     inherited Destroy;
end;

function TGBufferStream.FillBuffer: Integer;
begin
     result := 0;
     if not IsEOF then
     begin
          { Fill the buffer from the stream }
{$IFDEF CLR}
          Result := Stream.Read(FBuffer, 0, BufferSize);
          FBufferPtr := 0;
          FBufferEnd := result;
{$ELSE}
          Result := Stream.Read(Buffer^, BufferSize);
          FBufferPtr := Buffer;
          FBufferEnd := Buffer + result;
{$ENDIF}
     end;
     { If nothing was read, it must be the end of file. }
     if Result > 0 then
        FState := bsRead
     else
         FState := bsUnknown;
     if Assigned(fOnFillBuffer) then fOnFillBuffer(Self);
end;

function TGBufferStream.FlushBuffer: Boolean;
{ Write the output buffer to the stream }
{$IFDEF CLR}
var
   NumBytes: Integer;
   n: Integer;
begin
     { Determine the number of bytes in the buffer. }
     NumBytes := BufferPtr;
     { Write the buffer contents. }
     n := Stream.Write( Buffer, NumBytes );
     Result := NumBytes = n;
     { The buffer is empty, so reset the state. }
     FBufferPtr := 0;
{$ELSE}
var
   NumBytes: Integer;
begin
     { Determine the number of bytes in the buffer. }
     NumBytes := BufferPtr - Buffer;
     { Write the buffer contents. }
     Result := NumBytes = Stream.Write( Buffer^, NumBytes);
     { The buffer is empty, so reset the state. }
     FBufferPtr := Buffer;
{$ENDIF}
     FState := bsUnknown;
     if Assigned(fOnFlushBuffer) then FOnFlushBuffer(Self);
end;

{$IFNDEF CLR}
function TGBufferStream.Read(var Buffer; Count: Longint): Longint;
{ Read Count number of bytes.  If count is greater than buffersize, loop repeatedly }
var
   Ptr: TBufferIndexType;
   NumBytes: Integer;
begin
     if State = bsWrite then
        FlushBuffer
     else if BufferPtr = nil then
          FBufferPtr := BufferEnd;

     Ptr := @Buffer;
     Result := 0;
     while Count > 0 do
     begin
          { If the buffer is empty, then fill it. }
          if (BufferPtr = BufferEnd) and
             (FillBuffer = 0) then
             Break;
          NumBytes := BufferEnd - BufferPtr;
          if Count < NumBytes then
             NumBytes := Count;

          { Copy the buffer to the user's memory. }
          Move(BufferPtr^, Ptr^, NumBytes);
          Dec(Count, NumBytes);
          Inc(FBufferPtr, NumBytes);
          Inc(Result, NumBytes);
          Inc(Ptr, NumBytes);
     end;
end;

function TGBufferStream.Write(const Buffer; Count: Longint): Longint;
{ Write Count bytes from Buffer to the stream }
var
   Ptr: Pointer;
   NumBytes: Integer;
begin
     { If the stream is for reading, then ignore the current buffer
       by forcing the position of the underlying stream to match
       the buffered stream's position. }
     if State = bsRead then
        FStream.Position := Position
     else if BufferPtr = nil then
          PrepareBuffer;

     Ptr := @Buffer;
     Result := 0;                   { Total number of bytes written. }
     while Count > 0 do
     begin
          { Calculate the number of bytes remaining in the buffer. }
          NumBytes := BufferEnd - BufferPtr;
          if Count < NumBytes then
             NumBytes := Count;
          { Copy from the user's memory to the buffer. }
          Move(Ptr^, BufferPtr^, NumBytes);
          Dec(Count, NumBytes);
          Inc(FBufferPtr, NumBytes);
          Inc(Result, NumBytes);
          Inc(TRSNativeInt(Ptr), NumBytes);
          if (BufferPtr = BufferEnd) and (not FlushBuffer) then Break;
     end;
     { If anything remains in the buffer, then set the state to bsWrite. }
     if BufferPtr <> fBuffer then
        FState := bsWrite;
end;

function TGBufferStream.Seek(Offset: Longint; Origin: Word): Longint;
var
   CurrentPosition: Integer;
begin
     { Determine the current position. }
     CurrentPosition := Stream.Position + BufferPosition;

     { Determine the new position }
     case Origin of
          soFromBeginning: Result := Offset;
          soFromCurrent:   Result := Stream.Position + BufferPosition + Offset;
          soFromEnd:       Result := Stream.Size - Offset;
     else
         raise EStreamError.CreateFmt(SInvalidOrigin, [Origin]);
     end;

     { Is the desired position different from the current position? }
     if Result <> CurrentPosition then
     begin
          { Flush a partial write. }
          if (State = bsWrite) and not FlushBuffer then
             raise EStreamError.Create(SSeekError);
          { Reset the stream. }
          Stream.Position := Result;
          { Discard the current buffer. }
          FBufferPtr := nil;
          FState := bsUnknown;
     end;
end;

{$ELSE}

function TGBufferStream.Write(const Buffer: array of Byte; Offset,
  Count: Integer): Longint;
{ Write Count bytes from Buffer to the stream }
var
   NumBytes: Integer;
begin
     { If the stream is for reading, then ignore the current buffer
       by forcing the position of the underlying stream to match
       the buffered stream's position. }
     if State = bsRead then
        FStream.Position := Position
     else if BufferPtr = -1 then
          PrepareBuffer;

     Result := 0;                   { Total number of bytes written. }
     while Count > 0 do
     begin
          { Calculate the number of bytes remaining in the buffer. }
          NumBytes := BufferEnd - BufferPtr;
          if Count < NumBytes then
             NumBytes := Count;
          { Copy from the user's memory to the buffer. }
          System.Array.Copy(Buffer, Offset, FBuffer, BufferPtr, NumBytes);
//          Move(Ptr^, BufferPtr^, NumBytes);
          Dec(Count, NumBytes);
          Inc(FBufferPtr, NumBytes);
          Inc(Result, NumBytes);
          Inc(Offset, NumBytes);
          if (BufferPtr = BufferEnd) and (not FlushBuffer) then Break;
     end;
     { If anything remains in the buffer, then set the state to bsWrite. }
     if BufferPtr <> 0 then
        FState := bsWrite;
end;

function TGBufferStream.Seek(const Offset: Int64; Origin: TSeekOrigin): Int64;
var
   CurrentPosition: Integer;
begin
     { Determine the current position. }
     CurrentPosition := Stream.Position + BufferPosition;

     { Determine the new position }
     case Origin of
          soFromBeginning: Result := Offset;
          soFromCurrent:   Result := Stream.Position + BufferPosition + Offset;
          soFromEnd:       Result := Stream.Size - Offset;
     else
         raise EStreamError.CreateFmt(SInvalidOrigin, [Origin]);
     end;

     { Is the desired position different from the current position? }
     if Result <> CurrentPosition then
     begin
          { Flush a partial write. }
          if (State = bsWrite) and not FlushBuffer then
             raise EStreamError.Create(SSeekError);
          { Reset the stream. }
          Stream.Position := Result;
          { Discard the current buffer. }
          FBufferPtr := -1;
          FState := bsUnknown;
     end;
end;

function TGBufferStream.Read(var Buffer: array of Byte; Offset,
  Count: Integer): Longint;
{ Read Count number of bytes.  If count is greater than buffersize, loop repeatedly }
var
   NumBytes: Integer;
begin
     if State = bsWrite then
        FlushBuffer
     else if BufferPtr = -1 then
          FBufferPtr := BufferEnd;

     Result := 0;
     while Count > 0 do
     begin
          { If the buffer is empty, then fill it. }
          if (BufferPtr = BufferEnd) and
             (FillBuffer = 0) then Break;
          NumBytes := BufferEnd - BufferPtr;
          if Count < NumBytes then
             NumBytes := Count;

          { Copy the buffer to the user's memory. }
          System.Array.Copy(FBuffer, BufferPtr, Buffer, Offset, NumBytes);
          //Move(BufferPtr^, Ptr^, NumBytes);
          Dec(Count, NumBytes);
          Inc(FBufferPtr, NumBytes);
          Inc(Result, NumBytes);
          Inc(Offset, NumBytes);
     end;
end;
{$ENDIF}

procedure TGBufferStream.SetBufferSize(const Value: Cardinal);
{ Set the new buffer size }
begin
     if Value <> FBufferSize then
     begin
          if Value = 0 then
             raise EStreamError.CreateFmt( SInvalidBufferSize, [IntToStr(Value)] );
          FlushBuffer;
          FBufferSize := Value;
     end;
end;

function TGBufferStream.GetBufferPosition: Integer;
begin
     case State of
          bsUnknown: Result := 0;
          bsRead:    Result := BufferPtr - BufferEnd;
          bsWrite:   Result := BufferPtr {$IFNDEF CLR}- Buffer{$ENDIF};
     else
         raise EStreamError.Create(SInvalidBufferState);
     end;
end;

procedure TGBufferStream.PutBack(Ch: Byte);
{ Push a byte back onto the input buffer. }
begin
{$IFDEF CLR}
     if FBufferPtr <= 0 then
        raise EStreamError.Create(SPutBackOverflow);
     Dec(FBufferPtr);
     Buffer[BufferPtr] := TBufferInternalType(Ch);
{$ELSE}
     if FBufferPtr <= FBuffer then
        raise EStreamError.Create(SPutBackOverflow);
     Dec(FBufferPtr);
     BufferPtr[0] := TBufferInternalType(Ch);
{$ENDIF}
end;

function TGBufferStream.PrepareBuffer: Boolean;
begin
     { Unknown state, so start with an empty buffer. }
{$IFDEF CLR}
     FBufferPtr := 0;
     FBufferEnd := BufferSize;
{$ELSE}
     FBufferPtr := FBuffer;
     FBufferEnd := FBuffer + BufferSize;
{$ENDIF}
     result := True;
end;

procedure TGBufferStream.PutBack(Ch: Char);
{ Push a character back onto the input buffer. }
begin
{$IFDEF CLR}
     if FBufferPtr <= 0 then
        raise EStreamError.Create(SPutBackOverflow);
     Dec(FBufferPtr, SizeOf(Char));
     Buffer[BufferPtr] := TBufferInternalType(Ch);
{$ELSE}
     if FBufferPtr <= FBuffer then
        raise EStreamError.Create(SPutBackOverflow);
     Dec(FBufferPtr, SizeOf(Char));
     { Copy the bytes of the char back }
     Move(Ch, BufferPtr^, SizeOf(Char));
//     BufferPtr[0] := Ch;
{$ENDIF}
end;

function TGBufferStream.IsEof: Boolean;
begin
     Result := (BufferPtr = BufferEnd) and (Stream.Position = Stream.Size);
end;

{$IFDEF CLR}
procedure TGBufferStream.SetSize(NewSize: Int64);
begin
     // do nothing
end;
{$ENDIF}

procedure TGBufferStream.Close;
begin
     if State = bsWrite then
        FlushBuffer;
end;

{ TGTextStream }

constructor TGTextStream.Create(Stream: TStream{$IFDEF DELPHI12PLUS}; Encoding: TEncoding{$ENDIF});
begin
     inherited Create( Stream );
     FDelimiters := WhiteSpaceDelimiters;
{$IFDEF DELPHI12PLUS}
     FEncoding := Encoding;
     FPreambleStreamed := FEncoding <> nil;
{$ENDIF}
end;

constructor TGTextStream.Create(const Filename: String; Mode: Word{$IFDEF DELPHI12PLUS}; Encoding: TEncoding{$ENDIF});
begin
     inherited Create( TFileStream.Create(Filename, Mode) );
     FDelimiters := WhiteSpaceDelimiters;
     FOwnStream := True;
{$IFDEF DELPHI12PLUS}
     FEncoding := Encoding;
     FPreambleStreamed := FEncoding <> nil;
{$ENDIF}
end;

destructor TGTextStream.Destroy;
begin
     { Flush the buffer first and then free it }
     inherited Destroy;
     FreeStream;
end;

function TGTextStream.FillBuffer: Integer;
{$IFDEF DELPHI12PLUS}
var
   ASize: Integer;
{$ENDIF}
begin
     result := inherited FillBuffer;
     if (result > 0) and (not PreambleStreamed) then
     begin
          {$IFDEF DELPHI12PLUS}
          if (Encoding = nil) and (Stream is TStringStream) then
             FEncoding := TStringStream(Stream).Encoding;
          ASize := TEncoding.GetBufferEncoding(TBytes(BufferPtr), FEncoding);
          if ASize > Integer(BufferSize) then
             raise EStreamError.Create(SPreamblePastEof);
          Inc(FBufferPtr, ASize);
          {$ELSE}
          Inc(FBufferPtr, PreambleSize);
          {$ENDIF}
          PreambleStreamed := True;
     end;
end;

procedure TGTextStream.FreeStream;
begin
     if OwnStream then
     begin
          FStream.Free;
          FStream := nil;
     end;
end;

function TGTextStream.GetToken: String;
{ Read a token, delimited by arbitrary characters }
var
   Ch: Char;
begin
     Result := '';
     { Read the input one character at a time. }
     while GetChar(Ch) do
     begin
          { Check for delimiters. When a delimiter is found, the delimiter
            character is pushed back onto the input buffer before exiting
            from the loop. }
          if Pos(Ch, Delimiters) > 0 then
          begin
               Putback(Ch);
               Break;
          end;
          { Append to the token String }
          result := result + Ch;
         // AppendStr(Result, Ch);
     end;
end;

function TGTextStream.GetToken(const Delimiters: String; RestoreOld: Boolean): String;
{ Restore old indicates if current delimiters should be saved }
var
   HoldDelimiters: String;
begin
     if RestoreOld then
        HoldDelimiters := Self.Delimiters;
     Self.Delimiters := Delimiters;
     result := GetToken;
     if RestoreOld then
        Self.FDelimiters := HoldDelimiters;
end;

function TGTextStream.GetLine: String;
{ Read a line of text, stripping the end of line }
var
   Ch: Char;
begin
     Result := '';
     while GetChar(Ch)  and (not ((Ch = #10) or (Ch = #13))) do
          result := result + Ch;
     { If the end-of-line is CR, look for a subsequent LF. }
     if Ch = #13 then
        if GetChar(Ch) and (Ch <> #10) then
           Putback(Ch);
end;

procedure TGTextStream.SkipSpaces;
var
   C: Char;
begin
     while GetChar(C) do
     if C > ' ' then
     begin
          { Stop with the first nonspace character. }
          Putback(C);
          Break;
     end;
end;

function TGTextStream.SkipInLineSpaces: Boolean;
{ Skip only spaces in the current line, return whether the end of line was reached }
var
   C: Char;
begin
     result := False;
     while GetChar(C) do
     if C > ' ' then
     begin
          { Stop with the first nonspace character. }
          Putback(C);
          Break;
     end
     else if C = #13 then
     begin
          result := True;
          { Eat the CRLF }
          if GetChar(C) and (C <> #10) then
             Putback(C);
          Break;
     end
     else if C = #10 then
     begin
          result := True;
          Break;
     end;
end;

function TGTextStream.GetChar: Char;
{$IFDEF DELPHI12PLUS}
var
   Bytes: TBytes;
   BytePtr: TBufferType;
   Size: Integer;
   Chars: TCharArray;
{$ENDIF}
begin
     result := #0;          // initialize result, otherwise, at end of stream the result
                            // will NOT equal #0 as it should
{$IFDEF DELPHI12PLUS}
     repeat
        SetLength(Bytes,Length(Bytes)+1);
        BytePtr := @Bytes[Length(Bytes)-1];
        Size := Read(BytePtr^, 1);
        if Size = 0 then
           Exit;
     until Encoding.GetCharCount(Bytes) = 1;
     if Size > 0 then
     begin
          Chars := Encoding.GetChars(Bytes);
          result := Chars[0];
     end;
{$ELSE}
     Read(result, SizeOf(Char));
{$ENDIF}
end;

function TGTextStream.GetChar( var AChar: Char ): Boolean;
begin
     AChar := GetChar;
     result := AChar <> #0;
end;

function TGTextStream.GetInteger: Integer;
{ Read a token and convert it to an integer. }
begin
     SkipSpaces;
     Result := StrToInt(GetToken);
end;

function TGTextStream.GetFloat: Extended;
{ Read a token and convert it to a floating point number. }
begin
     SkipSpaces;
     Result := StrToFloat(GetToken);
end;

{$IFNDEF NEXTGEN}
function TGTextStream.Put(Ch: AnsiChar): TGTextStream;
begin
//     WriteBuffer(Ch, SizeOf(ch));
     Result := Put(String(Ch));
end;
{$ENDIF !NEXTGEN}

function TGTextStream.Put(WCh: WideChar): TGTextStream;
begin
//     WriteBuffer(WCh, SizeOf(WCh));
     Result := Put(String(WCh));
end;

function TGTextStream.Put(Int: Integer): TGTextStream;
begin
     Result := Put(IntToStr(Int));
end;

function TGTextStream.Put(Flt: Extended): TGTextStream;
begin
     Result := Put(FloatToStr(Flt));
end;

{$IFDEF DELPHI12PLUS}

function TGTextStream.PrepareBuffer: Boolean;
var
   Preamble: TBytes;
begin
     result := inherited PrepareBuffer;
     // Now, write out encoding
     if Encoding = nil then
     begin
          FEncoding := TEncoding.Default;
          PreambleStreamed := False;
     end;
     if not PreambleStreamed then
     begin
          Preamble := Encoding.GetPreamble;
          if Length(Preamble) > 0 then
             WriteBuffer(Preamble[0], Length(Preamble));
          PreambleStreamed := True;
     end;
end;

procedure TGTextStream.PutBack(Ch: Char);
var
   Bytes: TBytes;
   i: Integer;
begin
     if Encoding = nil then
        inherited PutBack(Ch)
     else
     begin
          Bytes := Encoding.GetBytes(String(Ch));
          for I := 1 to Length(Bytes) do
              PutBack(Bytes[Length(Bytes)-i]);
     end;
end;

{$ENDIF}

function TGTextStream.Put(const Str: String): TGTextStream;
{$IFDEF CLR}
var
   ch: Char;
begin
     for ch in Str do
         WriteBuffer(ch);
{$ELSE}
{$IFDEF DELPHI12PLUS}
var
   StrBuffer: TBytes;
begin
     if Encoding = nil then
        FEncoding := TEncoding.Default;
     StrBuffer := Encoding.GetBytes(Str);
     WriteBuffer(StrBuffer[0], Length(StrBuffer));
{$ELSE}
begin
     WriteBuffer(Pointer(Str)^, Length(Str)*SizeOf(Char));
{$ENDIF}
{$ENDIF}
     Result := Self;
end;

function TGTextStream.PutLine(const Str: String): TGTextStream;
begin
     Put(Str);
     Result := PutEOL;
end;

function TGTextStream.PutSpace: TGTextStream;
begin
     Result := Put(' ');
end;

function TGTextStream.PutTab: TGTextStream;
begin
     Result := Put(#9);
end;

function TGTextStream.PutEOL: TGTextStream;
begin
     Put(#13);
     Result := Put(#10);
end;

procedure TGTextStream.WriteArgs(Args: array of const);
{ Write arbitrary arguments, using default formatting. }
var
  I: Integer;
begin
    for I := Low(Args) to High(Args) do
    begin
{$IFDEF CLR}
        Put(String(Args[i]));
        if (I < High(Args)) and (not VarIsType(Variant(Args[I]), varChar)) then
           PutSpace;
{$ELSE}
        case Args[I].VType of
             vtInteger:         Put(Args[I].VInteger);
             vtBoolean:
             if Args[I].VBoolean then
                Put('True')
             else
                 Put('False');
             vtExtended:        Put(Args[I].VInteger);
{$IFNDEF NEXTGEN}
             vtChar:            Put(Char(Args[I].VChar));
             vtString:          Put(String(AnsiString(Args[I].VString^)));
             vtAnsiString:      Put(String(Args[I].VAnsiString));
             vtWideString:      Put(WideString(Args[i].VWideString^));
             vtPChar:           {$IFDEF UNICODE}Put(WideChar(Args[I].VPChar));{$ELSE}Put(Args[I].VPChar);{$ENDIF}
{$ENDIF !NEXTGEN}
             vtPointer:         Format('%p', [Args[I].VPointer]);
             vtClass:           Put(Args[I].VClass.ClassName);
             vtObject:
             begin
                  Put('(');
                  Put(TObject(Args[I].VObject).ClassName);
                  Put(')');
             end;
             vtWideChar:        Put(Args[I].VWideChar);
             vtCurrency:        Put(Args[I].VCurrency^);
             vtVariant:         Put(String(Args[I].VVariant^));
             vtInt64:           Put(IntToStr(Args[I].VInt64^));
{$IFDEF DELPHI12PLUS}
             vtUnicodeString:   Put(String(Args[I].VUnicodeString^));
{$ENDIF}
        end;
{$IFNDEF NEXTGEN}
        if (I < High(Args)) and (Args[I].VType <> vtChar) then
           PutSpace;
{$ENDIF !NEXTGEN}
{$ENDIF}
    end;
end;

procedure TGTextStream.WriteLn(Args: array of const);
begin
     WriteArgs(Args);
     PutEOL;
end;

procedure TGTextStream.Format(const Fmt: String; Args: array of const);
begin
     Put(SysUtils.Format(Fmt, Args));
end;

procedure TGTextStream.FormatLn(const Fmt: String; Args: array of const);
begin
     Put(SysUtils.Format(Fmt, Args));
     PutEOL;
end;

procedure TGTextStream.SetDelimiters(const Value: String);
begin
     if FDelimiters <> Value then
     begin
          FDelimiters := Value;
          if Pos( ' ', FDelimiters ) = 0 then
             FDelimiters := FDelimiters + ' ';
     end;
end;

function TGTextStream.PreambleSize: Integer;
var
   Preamble: TBytes;
begin
     SetLength(Preamble,0);
{$IFDEF DELPHI12PLUS}
     if Encoding = nil then
        FEncoding := TEncoding.Default;
     Preamble := Encoding.GetPreamble;
{$ENDIF}
     result := Length(Preamble);
end;

{$IFNDEF CLR}
function TGTextStream.Seek(Offset: Longint; Origin: Word): Longint;
begin
     result := inherited Seek(Offset, Origin);
     PreambleStreamed := result > 0;
end;

function TGTextStream.SeekExcludePreamble(Offset: Longint; Origin: Word): Longint;
begin
     result := inherited Seek(Offset, Origin);
     // if we have already written the preamble, modify the seek to exclude the
     // bytes of the preamble
     if PreambleStreamed and (PreambleSize > 0) then
        result := inherited Seek(PreambleSize, soFromCurrent);
end;
{$ENDIF}

{ TBaseStreamToken }

procedure TBaseStreamToken.Assign(Source: TPersistent);
begin
     if Source is TBaseStreamToken then
     begin
          FDataType := TBaseStreamToken(Source).DataType;
          FLineNumber := TBaseStreamToken(Source).LineNumber;
          FNumber := TBaseStreamToken(Source).Number;
          FText := TBaseStreamToken(Source).Text;
          FTag := TBaseStreamToken(Source).Tag;
          Exit;
     end;
     inherited Assign( Source );
end;

procedure TBaseStreamToken.Assign(const AValue: IInterface);
begin
     with AValue as IBaseStreamToken do
     begin
          Self.FDataType := DataType;
          Self.FLineNumber := LineNumber;
          Self.FNumber := Number;
          Self.FText := Text;
          Self.FTag := Tag;
     end;
end;

function TBaseStreamToken.Clone: IInterface;
begin
     result := TBaseStreamToken(ClonePersistentClass(Self, ClassType)) as IInterface;
end;

function TBaseStreamToken.GetDataType: Integer;
begin
     result := FDataType;
end;

function TBaseStreamToken.GetLineNumber: Integer;
begin
     result := FLineNumber;
end;

function TBaseStreamToken.GetNumber: Double;
begin
     result := FNumber;
end;

function TBaseStreamToken.GetText: String;
begin
     result := FText;
end;

function TBaseStreamToken.ToString: String;
begin
     case DataType of
         Ord(sttString):        result := SEscapeChar + SStringChar + Text +
                                          SStringChar + SEscapeChar;
         Ord(sttInteger):       result := IntToStr(Trunc(Number));
         Ord(sttFloat):         result := Format(FloatFormatStr,[Number]);
         Ord(sttVariable):      result := '?'+Text;
     else
         if Text <> '' then
            result := Text
         else
             result := Chr(DataType);
     end;
end;

procedure TBaseStreamToken.SetDataType(const Value: Integer);
begin
     FDataType := Value;
end;

procedure TBaseStreamToken.SetLineNumber(const Value: Integer);
begin
     FLineNumber := Value;
end;

procedure TBaseStreamToken.SetNumber(const Value: Double);
begin
     FNumber := Value;
end;

procedure TBaseStreamToken.SetText(const Value: String);
begin
     FText := Value;
end;

function TBaseStreamToken.Equals(AValue: TPersistent): Boolean;
var
   AStreamToken: IBaseStreamToken;
begin
     if Supports(AValue, IBaseStreamToken, AStreamToken) then
        result := Equals( AStreamToken )
     else
         result := False;
end;

function TBaseStreamToken.Equals(const AValue: IInterface): Boolean;
var
   AStreamToken: IBaseStreamToken;
begin
     result := Supports(AValue, IBaseStreamToken, AStreamToken);
     if result then
        result := (FDataType = AStreamToken.DataType) and
                  (FLineNumber = AStreamToken.LineNumber) and
                  (FNumber = AStreamToken.Number);
end;

function TBaseStreamToken.GetTag: TRSTag;
begin
     result := FTag;
end;

procedure TBaseStreamToken.SetTag(const Value: TRSTag);
begin
     FTag := Value;
end;

function TBaseStreamToken._AddRef: Integer;
begin
{$IFDEF DELPHI18PLUS}
  {$IFNDEF AUTOREFCOUNT}
    Result := AtomicIncrement(FRefCount);
  {$ELSE}
    Result := __ObjAddRef;
  {$ENDIF}
{$ELSE}
  {$IFNDEF DELPHI16PLUS}
  result := InterlockedIncrement(FRefCount);
  {$ELSE}
  result := TInterlocked.Increment(FRefCount);
  {$ENDIF}
{$ENDIF}
end;

function TBaseStreamToken._Release: Integer;
begin
{$IFDEF DELPHI18PLUS}
  {$IFNDEF AUTOREFCOUNT}
    Result := AtomicDecrement(FRefCount);
    if Result = 0 then
      Destroy;
  {$ELSE}
    Result := __ObjRelease;
  {$ENDIF}
{$ELSE}
  {$IFNDEF DELPHI16PLUS}
  result := InterlockedDecrement(FRefCount);
  {$ELSE}
  result := TInterlocked.Decrement(FRefCount);
  {$ENDIF}
  if result = 0 then
     Destroy;
{$ENDIF}
end;

function TBaseStreamToken.QueryInterface(const IID: TGUID; out Obj): HResult;
begin
  if GetInterface(IID, Obj) then Result := 0 else Result := E_NOINTERFACE;
end;

function TBaseStreamToken.GetWhitespace: String;
begin
     result := FWhitespace;
end;

procedure TBaseStreamToken.SetWhitespace(const Value: String);
begin
     FWhitespace := Value;     
end;

{ TBaseStreamTokenList Methods }

function TBaseStreamTokenList.Add(const Item: IBaseStreamToken): Integer;
begin
     result := inherited Add( Item );
end;

function TBaseStreamTokenList.Equals(const AList: IBaseStreamTokenList): Boolean;
var
   i: Integer;
begin
    result := False;
    if ValueCount <> AList.ValueCount then Exit;
    for i := 0 to AList.ValueCount - 1 do
       if AList[i] <> nil then
       begin
            if Value[i] = nil then Exit
            else if not Value[i].Equals(AList[i]) then Exit;
       end
       else if Value[i] <> nil then Exit;
    result := True;
end;

function TBaseStreamTokenList.Equals(AList: TObject): Boolean;
var
   i: Integer;
   Val: IBaseStreamToken;
begin
     result := AList = Self;
     if result then Exit;
     if AList is TInterfaceList then
     with TInterfaceList(AList) do
     begin
         result := False;
         if Self.ValueCount <> Count then Exit;
         for i := 0 to Self.ValueCount - 1 do
            if Items[i] <> nil then
            begin
                 if (Self.Value[i] = nil) or
                    (not Supports(Items[i], IBaseStreamToken, Val) ) then Exit
                 else if not Self.Value[i].Equals(Val) then Exit;
            end
            else if Self.Value[i] <> nil then Exit;
         result := True;
     end
     else
         result := False;
end;

function TBaseStreamTokenList.First: IBaseStreamToken;
begin
     result := inherited First as IBaseStreamToken;
end;

function TBaseStreamTokenList.Get(Index: Integer): IBaseStreamToken;
begin
     result := inherited Get(Index) as IBaseStreamToken;
end;

function TBaseStreamTokenList.GetValueCount: Integer;
begin
     result := Count;
end;

function TBaseStreamTokenList.IndexOf(const Item: IBaseStreamToken): Integer;
begin
     result := inherited IndexOf(Item);
end;

procedure TBaseStreamTokenList.Insert(Index: Integer; const Item: IBaseStreamToken);
begin
     inherited Insert( Index, Item );
end;

function TBaseStreamTokenList.Last: IBaseStreamToken;
begin
     result := inherited Last as IBaseStreamToken;
end;

procedure TBaseStreamTokenList.Put(Index: Integer; const Value: IBaseStreamToken);
begin
     inherited Put( Index, Value );
end;

function TBaseStreamTokenList.Remove(const Item: IBaseStreamToken): Integer;
begin
     result := inherited Remove(Item);
end;

function TBaseStreamTokenList.Clone: IInterface;
var
   Clone: TBaseStreamTokenList;
begin
     Clone := TBaseStreamTokenList.Create;
     result := Clone;
     Clone.Assign( Self as IInterface );
end;

procedure TBaseStreamTokenList.Assign(Source: TPersistent);
begin
     raise Exception.CreateFmt('Error assigning list',[ClassName, Source.ClassName]);
end;

procedure TBaseStreamTokenList.Assign(const AValue: IInterface);
var
   DValue: IBaseStreamTokenList;
   i: Integer;
begin
     DValue := AValue as IBaseStreamTokenList;
     Clear;
     for i := 0 to DValue.ValueCount - 1 do
         Add( DValue[i].Clone as IBaseStreamToken );
     FTag := DValue.Tag;
end;

function TBaseStreamTokenList.ToString: String;
var
   i: Integer;
begin
     result := '';
     for i := 0 to Count - 1 do
         result := result + Value[i].ToString + ' ';
end;

procedure TBaseStreamTokenList.Append(const List: IBaseStreamTokenList);
var
   i: Integer;
begin
     for i := 0 to List.ValueCount - 1 do
         Add( List[i] );
end;

function TBaseStreamTokenList.Equals(const AValue: IInterface): Boolean;
var
   VV: IBaseStreamTokenList;
begin
     result := Supports(AValue, IBaseStreamTokenList, VV );
     if result then
        result := Equals( VV );
end;

function TBaseStreamTokenList.GetTag: TRSTag;
begin
     result := FTag;
end;

procedure TBaseStreamTokenList.SetTag(const Value: TRSTag);
begin
     FTag := Value;
end;

{ TGSpecialFiler }

{$IFNDEF CLR}
type
    TWriterCrack = class(TWriter);
    TReaderCrack = class(TReader);

constructor TGSpecialFiler.Create(Writer: TWriter);
begin
     inherited Create;
     FWriter := Writer;
end;

constructor TGSpecialFiler.Create;
begin
     inherited Create;
end;

constructor TGSpecialFiler.Create(Reader: TReader);
begin
     inherited Create;
     FReader := Reader;
end;

procedure TGSpecialFiler.ReadCollection(Collection: TCollection);
var
  Item: TPersistent;
begin
     if Reader = nil then
        raise EStreamError.Create( SNoReaderDefined );
     Collection.BeginUpdate;
     try
        Collection.Clear;
        Reader.ReadValue;
        while not Reader.EndOfList do
        begin
             // read the class name
             Reader.ReadListBegin;
             Reader.ReadStr; // HiddenItemClass
             Item := TCollectionItemClass(FindClass( Reader.ReadString )).Create(Collection);
             while not Reader.EndOfList do TReaderCrack(Reader).ReadProperty(Item);
             Reader.ReadListEnd;
        end;
        Reader.ReadListEnd;
     finally
        Collection.EndUpdate;
     end;
end;

procedure TGSpecialFiler.WriteCollection(Value: TCollection);
var
   i: Integer;
begin
     if Writer = nil then
        raise EStreamError.Create( SNoWriterDefined );
     TWriterCrack(Writer).WriteValue(vaCollection);
     if Value <> nil then
        for I := 0 to Value.Count - 1 do
        begin
             Writer.WriteListBegin;
             TWriterCrack(Writer).WritePropName('HiddenItemClass');
             Writer.WriteString(Value.Items[i].ClassName);
             TWriterCrack(Writer).WriteProperties(Value.Items[I]);
             Writer.WriteListEnd;
        end;
     Writer.WriteListEnd;
end;
{$ENDIF}
end.

